import 'dart:async';
import 'dart:io';
import 'package:heavylist/heavylist.dart';
import 'package:jsonstreamreader/abstractreader.dart';

import 'complexexpression.dart';
import 'jsontoken.dart';
import 'jsontokenizer.dart';
import 'source/remote/remotesource.dart';
import 'state/globalreadermanager.dart';
import 'state/globalreaderstatemanager.dart';
import 'state/reader/readerstate.dart';
import 'streamer.dart';

///Reads a List of JsonTokens and calls the relevant callbacks
class Reader extends AbstractReader {
  JsonTokenizer tokenizer;
  Function _trailing;
  List<MapEntry<ComplexExpression, Function>> _filters;
  Function _done;
  Function _fail;
  Function _progress;
  Function _onpause;
  Duration _delay;
  int _combine = 0;
  ReaderState _state;
  bool _paused = false;
  bool _running = false;
  bool _hault = false;
  String _reference = 'N';
  HeavyList<MapEntry<String, dynamic>> _list;
  StreamSubscription<List<MapEntry<String, dynamic>>> _subscribe;
  StreamController _parsing = new StreamController();
  Reader.fromState(ReaderState state) {
    GlobalReaderManager.addReader(this);
    _reference = 'S';
    this._delay = new Duration(microseconds: state.delay);
    this.tokenizer = new JsonTokenizer.fromState(
        new Streamer(new File('/empty')), state.group,
        combine: 1);
    //recreate JsonTokens
    JsonToken.fromState(state.group, state.jsontokenstate);
    _state = state;
    new Timer(new Duration(microseconds: 10), () {
      if (!_hault) {
        _stateFire();
      }
    });
  }
  void _stateFire() {
    this._parse(_state.parse).then((String d) {
      new Timer(_delay, () {
        this._partial(_state);
      });
    }).catchError((err) {
      if (this._fail != null) {
        _parsing.close();
        this._fail(err);
      }
    });
  }

  void _partial(ReaderState state) {
    Streamer stream;
    new Timer(new Duration(microseconds: 1), () {
      //parse state.parse if>0 and then run below
      if (state.isRemote) {
        RemoteSource r = new RemoteSource.fromString(state.file);
        stream = new Streamer(r,
            start: state.start, end: state.end, chunksize: state.chunksize);
        new Timer(new Duration(microseconds: 0), () {
          r.download();
        });
      } else {
        File r = new File(state.file);
        stream = new Streamer(r,
            start: state.start, end: state.end, chunksize: state.chunksize);
      }
      this._combine = state.combine;
      this._delay = new Duration(microseconds: state.delay);
      this._filters = new List<MapEntry<ComplexExpression, Function>>();
      this.tokenizer = new JsonTokenizer.fromState(stream, state.group,
          combine: this._combine);
      _subscribe = JsonToken.event(this.tokenizer.identity).listen(
          (List<MapEntry<String, dynamic>> data) {
        try {
          this._parse(data).then((String d) {
            new Timer(_delay, () {
              this._run();
            });
          }).catchError((err) {
            if (this._fail != null) {
              _parsing.close();
              this._fail(err);
            }
          });
        } catch (e) {
          if (this._fail != null) {
            _parsing.close();
            this._fail(e);
          }
        }
      }, onError: (error) {
        if (this._fail != null) {
          _parsing.close();
          this._fail(error);
        }
      }, onDone: () {
        _parsing.close();
        if (this._progress != null) {
          this._progress(100.0);
        }
        if (this._done != null) {
          this._done();
        }
      });

      new Timer(new Duration(microseconds: 0), () {
        this._run();
      });
    });
  }

  Reader(Streamer stream, {int delay = 0, combine = 0}) {
    GlobalReaderManager.addReader(this);
    this._combine = combine;
    this._delay = new Duration(microseconds: delay);
    this._filters = new List<MapEntry<ComplexExpression, Function>>();
    this.tokenizer = new JsonTokenizer(stream, combine: this._combine);

    _subscribe = JsonToken.event(this.tokenizer.identity).listen(
        (List<MapEntry<String, dynamic>> data) {
      try {
        this._parse(data).then((String d) {
          new Timer(_delay, () {
            this._run();
          });
        }).catchError((err) {
          if (this._fail != null) {
            _parsing.close();
            this._fail(err);
          }
        });
      } catch (e) {
        if (this._fail != null) {
          _parsing.close();
          this._fail(e);
        }
      }
    }, onError: (error) {
      if (this._fail != null) {
        _parsing.close();
        this._fail(error);
      }
    }, onDone: () {
      if (this._progress != null) {
        this._progress(100.0);
      }
      if (this._done != null) {
        _parsing.close();
        this._done();
      }
    });
    new Timer(new Duration(microseconds: 0), () {
      if (!_hault) {
        _fire();
      }
    });
  }
  void _fire() {
    new Timer(new Duration(microseconds: 0), () {
      this._run();
    });
  }

  _run() {
    this.tokenizer.tokenize().then((int item) {
      if (JsonToken.getFired(this.tokenizer.identity) == 0 && item == 0) {
        new Timer(new Duration(microseconds: 0), () {
          this._run();
        });
      }
    }).catchError((err) {
      if (this._fail != null) {
        _parsing.close();
        this._fail(err);
      }
    });
  }

  Future<String> _parse(List<MapEntry<String, dynamic>> parse) async {
    //this.tokenizer.destroy();
    Completer<String> completer = new Completer<String>();
    this._list = new HeavyList<MapEntry<String, dynamic>>(parse);
    _parsing.add(null);
    _running = true;
    _list.loop(new Duration(microseconds: 0),
        (List<MapEntry<String, dynamic>> origin) {
      _list.empty();
      JsonToken.resetFired(this.tokenizer.identity);
      if (this._progress != null) {
        this._progress(Streamer.progess(this.tokenizer.identity));
      }
      completer.complete("done");
    }, (MapEntry<String, dynamic> item, Function resume) {
      dynamic firstValue = item.value;
      dynamic firstKey = item.key;
      List<String> ikeys = [];
      if (firstValue.runtimeType.toString().contains("HashMap")) {
        ikeys = item.value.keys.toList();
      }
      try {
        if (this._trailing != null) {
          var response = this._trailing(firstValue, firstKey);
          bool isFuture = response.runtimeType.toString().contains('Future');
          if (isFuture) {
            response.then((dynamic v) {
              HeavyList<MapEntry<ComplexExpression, Function>> sublist =
                  HeavyList<MapEntry<ComplexExpression, Function>>(
                      this._filters);
              sublist.loop(new Duration(microseconds: 0),
                  (List<MapEntry<ComplexExpression, Function>> suborigin) {
                resume();
              }, (MapEntry<ComplexExpression, Function> subitem,
                      Function subresume) {
                if (subitem.key.evaluate(ikeys, firstKey)) {
                  var subresponse = subitem.value(firstValue, firstKey);
                  bool subisFuture =
                      subresponse.runtimeType.toString().contains('Future');
                  if (subisFuture) {
                    subresponse.then((dynamic v) {
                      subresume();
                    }).catchError((error) {
                      completer.completeError(error);
                    });
                  } else {
                    subresume();
                  }
                } else {
                  subresume();
                }
              });
            }).catchError((error) {
              completer.completeError(error);
            });
          } else {
            _list.clearRegistered();
            resume();
          }
        } else if (this._filters.length > 0) {
          HeavyList<MapEntry<ComplexExpression, Function>> sublist =
              HeavyList<MapEntry<ComplexExpression, Function>>(this._filters);
          sublist.loop(new Duration(microseconds: 0),
              (List<MapEntry<ComplexExpression, Function>> suborigin) {
            resume();
          }, (MapEntry<ComplexExpression, Function> subitem,
                  Function subresume) {
            if (subitem.key.evaluate(ikeys, firstKey)) {
              var subresponse = subitem.value(firstValue, firstKey);
              bool subisFuture =
                  subresponse.runtimeType.toString().contains('Future');
              if (subisFuture) {
                subresponse.then((dynamic v) {
                  subresume();
                }).catchError((error) {
                  completer.completeError(error);
                });
              } else {
                subresume();
              }
            } else {
              subresume();
            }
          });
        } else {
          _list.clearRegistered();
          resume();
        }
      } catch (err) {
        completer.completeError(err);
      }
    });
    return completer.future;
  }

  Reader trailing(dynamic func(dynamic value, String key)) {
    this._trailing = func;
    return this;
  }

  Reader done(void func()) {
    this._done = func;
    return this;
  }

  Reader fail(void func(dynamic err)) {
    this._fail = func;
    return this;
  }

  Reader progress(void func(double progress)) {
    this._progress = func;
    return this;
  }

  //Prevent the reader from starting immediately
  void hault() {
    _hault = true;
  }

  //if the reader has hault,start now
  void begin() {
    if (_hault) {
      _hault = false;
      if (this._reference == 'N') {
        _fire();
      } else {
        _stateFire();
      }
    }
  }

  Reader filter(
      String expression, Future<dynamic> func(dynamic value, String key)) {
    this._filters.add(new MapEntry(new ComplexExpression(expression), func));
    return this;
  }

  Reader onpause(void func(ReaderState state)) {
    this._onpause = func;
    return this;
  }

  //Return the State Reader was in upon pausing
  Future<ReaderState> pause() async {
    Completer<ReaderState> completer = new Completer<ReaderState>();
    if (!this._paused) {
      if (_running) {
        new Timer(new Duration(microseconds: 0), () {
          this._pause();
          completer.complete(this._state);
        });
      } else {
        StreamSubscription sb;
        sb = _parsing.stream.listen((onData) {
          this._pause();
          sb.cancel();
          if (!completer.isCompleted) {
            completer.complete(this._state);
          }
        });
      }
    }
    return completer.future;
  }

  void _pause() {
    this._paused = true;
    _subscribe.pause();
    this._list.pause();
    this._state = new ReaderState(
        JsonToken.getState(this.tokenizer.identity),
        this.tokenizer.identity,
        this.tokenizer.stream.path,
        this.tokenizer.stream.start,
        this.tokenizer.stream.end,
        this._delay.inMicroseconds,
        this._combine,
        this.tokenizer.stream.chunksize,
        this._list.getList());
    if (this._onpause != null) {
      this._onpause(this._state);
    }
    GlobalReaderStateManager.addState(this._state);
  }

  void resume() {
    if (this._paused) {
      this._paused = false;
      this._list.resume();
      _subscribe.resume();
      GlobalReaderStateManager.removeState(this._state);
    }
  }
}
