import 'dart:async';
import 'dart:convert';
import 'dart:io';

import 'package:jsonstreamreader/source/remote/fileawaite.dart';

import 'remote/resourcestate.dart';
import 'remote/filestateprovider.dart';

///Normalizes a File and RemoteSource
class Normalized {
  var _file;
  String runtime;
  int _calls = 0;
  //last chunk fetch
  int _last = 0;
  int _length = 0;
  int _end = 0;
  bool _lengthFound = false;
  bool _reachedEndofStream = false;
  bool _endResolved = false;
  Normalized(var file) {
    this._file = file;
    this.runtime = file.runtimeType.toString();
  }
  //temp path of downloaded file
  String get path {
    return this._file.path;
  }

  //instance of File(dart.io)
  dynamic get file {
    return this._file;
  }

  //identify range where to start and stop reading files
  Normalized setBoundary(int start, int end) {
    this._end = end;
    this._last = start;
    return this;
  }

//delete a file
  Future<FileSystemEntity> delete() async {
    String runtime = this._file.runtimeType.toString();
    if (runtime != '_File') {
      return this._file.delete();
    } else {
      return this._file.getFile().delete();
    }
  }

//instance of File(dart.io)
  dynamic getFile() {
    return this._file;
  }

  Stream<String> _next(int chunk) {
    StreamController<String> c = new StreamController<String>();
    if (!this._lengthFound) {
      this._file.length().then((int l) {
        this._length = l;
        this._lengthFound = true;
        if (l == 0) {
          _irregular(chunk).listen((String line) {
            c.add(line);
          }, onDone: () {
            c.close();
          }, onError: (e) {
            c.addError(e);
          });
        } else {
          _regular(chunk, l).listen((String line) {
            c.add(line);
          }, onDone: () {
            c.close();
          }, onError: (e) {
            c.addError(e);
          });
        }
      }).catchError((error) {
        c.addError(error);
      });
    } else {
      if (this._length == 0) {
        _irregular(chunk).listen((String line) {
          c.add(line);
        }, onDone: () {
          c.close();
        }, onError: (e) {
          c.addError(e);
        });
      } else {
        _regular(chunk, this._length).listen((String line) {
          c.add(line);
        }, onDone: () {
          c.close();
        }, onError: (e) {
          c.addError(e);
        });
      }
    }
    return c.stream;
  }

  Stream<String> _regular(int chunk, int l) {
    StreamController<String> c = new StreamController<String>();
    int end = this._last + chunk;
    int endProcess = 0;
    if (this._end > 0 && end >= this._end) {
      end = this._end;
      endProcess = 1;
    }
    if (end > l && !_reachedEndofStream && !_endResolved) {
      _reachedEndofStream = true;
      this._file.openRead(this._last, l).transform(utf8.decoder).listen(
          (String line) {
        c.add(line);
      }, onDone: () {
        this._last = l;
        if (endProcess > 0) {
          this._endResolved = true;
        }
        c.close();
      }, onError: (e) {
        c.addError(e);
      });
    } else if (end < l && !_endResolved) {
      this._file.openRead(this._last, end).transform(utf8.decoder).listen(
          (String line) {
        c.add(line);
      }, onDone: () {
        this._last = end;
        if (endProcess > 0) {
          this._endResolved = true;
        }
        c.close();
      }, onError: (e) {
        c.addError(e);
      });
    } else if (end == l && !_endResolved) {
      this._file.openRead(this._last, end).transform(utf8.decoder).listen(
          (String line) {
        c.add(line);
      }, onDone: () {
        this._last = end;
        if (endProcess > 0) {
          this._endResolved = true;
        }
        c.close();
      }, onError: (e) {
        c.addError(e);
      });
    } else {
      new Timer(new Duration(microseconds: 0), () {
        c.add('');
        c.close();
      });
    }
    return c.stream;
  }

  Stream<String> _irregular(int chunk) {
    StreamController<String> c = new StreamController<String>();
    int end = (this._last + chunk);
    int endProcess = 0;
    if (this._end > 0 && end >= this._end) {
      end = this._end;
      endProcess = 1;
    }
    StreamSubscription<ResourceState> fstream;
    this._file.sp.pushState(end, FileStateProvider.CALLED);
    if (end < this._file.getCurrent() && !_endResolved) {
      this._file.openRead(this._last, end).transform(utf8.decoder).listen(
          (String line) {
        if (!c.isClosed) {
          c.add(line);
        } else {
          fstream.cancel();
        }
      }, onDone: () {
        this._last = end;
        if (endProcess > 0) {
          this._endResolved = true;
        }
        c.close();
      }, onError: (e) {
        c.addError(e);
      });
    } else if (end == this._file.getCurrent() && !_endResolved) {
      this._file.openRead(this._last, end).transform(utf8.decoder).listen(
          (String line) {
        if (!c.isClosed) {
          c.add(line);
        } else {
          fstream.cancel();
        }
      }, onDone: () {
        this._last = end;
        if (endProcess > 0) {
          this._endResolved = true;
        }
        c.close();
      }, onError: (e) {
        c.addError(e);
      });
    } else if (end > this._file.getCurrent() &&
        this._file.isCompleted() &&
        !this._reachedEndofStream &&
        !_endResolved) {
      this._reachedEndofStream = true;
      this._file.openRead(this._last).transform(utf8.decoder).listen(
          (String line) {
        if (!c.isClosed) {
          c.add(line);
        } else {
          fstream.cancel();
        }
      }, onDone: () {
        this._last = this._file.getCurrent();
        if (endProcess > 0) {
          this._endResolved = true;
        }
        c.close();
      }, onError: (e) {
        c.addError(e);
      });
    } else if ((end > this._file.getCurrent() &&
            this._file.isCompleted() &&
            this._reachedEndofStream) ||
        this._endResolved) {
      new Timer(new Duration(microseconds: 0), () {
        c.add('');
        c.close();
      });
    } else if (!this._endResolved && !this._reachedEndofStream) {
      if (endProcess == 1) {
        this._endResolved = true;
        this._last = end;
      }
      new FileAwaite(end, endProcess, this._last, this._file.sp,
          (FileAwaiteResponse response) {
        if (response.state == FileAwaiteState.COMPLETED) {
          this._reachedEndofStream = true;
        }
        if (response.value == 0) {
          this._file.openRead(response.last).transform(utf8.decoder).listen(
              (String line) {
            if (!c.isClosed) {
              c.add(line);
            } else {
              fstream.cancel();
            }
          }, onDone: () {
            this._last = this._file.getCurrent();
            c.close();
          }, onError: (e) {
            c.addError(e);
          });
        } else {
          this
              ._file
              .openRead(response.last, response.value)
              .transform(utf8.decoder)
              .listen((String line) {
            if (!c.isClosed) {
              c.add(line);
            } else {
              fstream.cancel();
            }
          }, onDone: () {
            this._last = response.value;
            c.close();
          }, onError: (e) {
            c.addError(e);
          });
        }
      });
    } else {
      new Timer(new Duration(microseconds: 0), () {
        if (!c.isClosed) {
          c.add('');
        }
        c.close();
      });
    }
    return c.stream;
  }

  //return last number of bytes loaded
  int getLast() {
    return this._last;
  }

//get progress of read
  double getProgress() {
    if (this._end > 0) {
      return this._last / this._end;
    } else if (this._length == 0 && this._file.isCompleted()) {
      double p = this._last / this._file.getCurrent();
      if (p < 0.4) {
        p = 0.4 + 0.4 / ((1 - p) * 10);
      }
      return p;
    } else if (this._length == 0 &&
        this._file.isCompleted() &&
        this._last == this._file.getCurrent()) {
      return 1;
    } else if (this._length == 0) {
      return 0.4 - (0.4 / this._calls);
    } else if (this._length == 0 && this._file.isCompleted()) {
      return this._last / this._file.getCurrent();
    } else {
      return this._last / this._length;
    }
  }

//get next range of bytes
  Stream<String> next(int chunk) {
    this._calls++;
    return this._next(chunk);
  }
}
