import 'filestateprovider.dart';
import 'resourcestate.dart';

///A safeguard to ensure the right bytes are returned
///when requesting a range of bytes from RemoteSource openRead
enum FileAwaiteState { COMPLETED, FOUND }

///Length Completed Response
class FileAwaiteResponse {
  int value;
  int last;
  FileAwaiteState state;
  int endProcess;
  FileAwaiteResponse(this.state, this.value, this.last, this.endProcess);
}

///Identify needed size and loaded size
class FileAwaite {
  int _end;
  int _endProcess;
  int _last;
  FileStateProvider _sp;
  bool _completed = false;
  Function _callback;
  FileAwaite(int end, int endProcess, int last, FileStateProvider sp,
      void callback(FileAwaiteResponse resp)) {
    this._end = end;
    this._sp = sp;
    this._callback = callback;
    this._endProcess = endProcess;
    this._last = last;
    this._sp.done.listen((d) {
      if (!_completed) {
        _completed = true;
        _callback(new FileAwaiteResponse(
            FileAwaiteState.COMPLETED, 0, this._last, _endProcess));
      }
    });
    this._sp.ready().listen((ResourceState data) {
      if (data.length >= _end) {
        if (!_completed) {
          _completed = true;
          _callback(new FileAwaiteResponse(
              FileAwaiteState.FOUND, data.length, this._last, _endProcess));
        }
      }
    });
  }
}
