import 'dart:async';

import 'package:heavylist/heavylist.dart';

import '../abstractreader.dart';
import 'reader/abstractstatereader.dart';

///Stores both Reader and StreamReader
class GlobalReaderManager {
  static List<AbstractReader> _readers = [];

  static void addReader(AbstractReader reader) {
    GlobalReaderManager._readers.add(reader);
  }

  static bool removeReader(AbstractReader reader) {
    return GlobalReaderManager._readers.remove(reader);
  }

  //if state is null pause all
  //if state is not null pause state
  static Future<List<AbstractStateReader>> pause(
      [AbstractReader reader]) async {
    List<AbstractStateReader> items = new List<AbstractStateReader>();
    Completer<List<AbstractStateReader>> completer =
        new Completer<List<AbstractStateReader>>();
    HeavyList<AbstractReader> list =
        new HeavyList<AbstractReader>(GlobalReaderManager._readers);
    if (reader == null) {
      list.loop(new Duration(microseconds: 0), (List<AbstractReader> origin) {
        completer.complete(items);
      }, (AbstractReader read, Function next) {
        read.pause().then((AbstractStateReader i) {
          items.add(i);
          next();
        }).catchError((error) {
          next();
        });
      });
    } else {
      return reader.pause().then((AbstractStateReader i) {
        items.add(i);
        completer.complete(items);
      }).catchError((error) {});
    }
    return completer.future;
  }

  //if state is null resume all
  //if state is not null resume state
  static void resume([AbstractReader reader]) {
    if (reader == null) {
      _readers.forEach((read) {
        read.resume();
      });
    } else {
      reader.resume();
    }
  }
}
