import 'dart:async';

import 'state/filestate.dart';

import 'source/normalized.dart';

///Identified chunk for a file based on
///chunksize
class Streamer {
  //chunk size
  int _size;
  //max size in kb of each chunk
  int _max;
  int _end;
  Normalized _file;
  static Map<int, double> _progress = new Map<int, double>();
  int group;

  ///
  /// @file file to be parsed
  /// @start where to start parsing
  /// @end where to stop parsing
  ///
  Streamer(var file, {start = 0, end = 0, chunksize = 100}) {
    String runtime = file.runtimeType.toString();
    if (runtime != '_File' && runtime != 'RemoteSource') {
      throw "Streamer accepts only File or RemoteSource";
    }
    this._file = Normalized(file);
    this._end = end;
    this._file.setBoundary(start, end);
    this._size = 0;
    FileState.addState(file);
    this._max = chunksize;
  }
  setGroup(int group) {
    this.group = group;
    Streamer._progress[this.group] = 0;
  }

  get end {
    return this._end;
  }

  get start {
    return this._file.getLast();
  }

  dynamic get file {
    return this._file.file;
  }

  dynamic get path {
    return this._file.path;
  }

  String get normalized {
    return this._file.file;
  }

  int get chunksize {
    return this._max;
  }

  Future<String> next() async {
    Completer<String> completer = new Completer<String>();
    if (this._size > 0) {
      this._serv().then((String value) {
        completer.complete(value);
      }).catchError((error) {
        completer.completeError(error);
      });
    } else {
      try {
        this._file.getFile().length().then((int value) {
          this._size = value;
          this._serv().then((String value) {
            completer.complete(value);
          }).catchError((error) {
            completer.completeError(error);
          });
        }).catchError((error) {
          completer.completeError(error);
        });
      } catch (e) {
        new Timer(new Duration(microseconds: 0), () {
          completer.completeError(e);
        });
      }
    }
    return completer.future;
  }

  Future<String> _serv() async {
    Completer<String> completer = new Completer<String>();
    StringBuffer sb = new StringBuffer();
    this._file.next(this._max * 1000).listen((String d) {
      Streamer._progress[this.group] = this._file.getProgress() * 100;
      sb.write(d);
    }, onDone: () {
      completer.complete(sb.toString());
      sb.clear();
    }, onError: (err) {
      completer.completeError(err);
    });
    return completer.future;
  }

  static double progess(group) {
    return Streamer._progress[group];
  }
}
